import threading
import logging
import time

logger = logging.getLogger(__name__)

class TService(threading.Thread):

    def __init__(self, name='service', interval=1E9):
        super().__init__(name=name, daemon=True)
        self._stop_event = threading.Event()
        self._update_event = threading.Event()
        self._update_lock = threading.Lock()
        self._update_messages = []
        self._interval = interval
        self._stopped = False
        self._has_started = False


    def set_interval(self, interval):
        self._interval = interval


    def run_start(self):
        pass


    def run_end(self):
        pass


    def run_loop(self):
        pass


    def run_update(self, message):
        pass


    def try_start(self):
        self._has_started = True
        try:
            self.run_start()
            return True
        except:
            logger.exception('Unexpected error in start')
            self._stopped = True
            return False


    def try_loop(self):
        try:
            self.run_loop()
            return True
        except:
            logger.exception('Unexpected error in loop')
            return False


    def try_update(self):
        try:
            with self._update_lock:
                if len(self._update_messages) > 0:
                    message = self._update_messages.pop()
                    self.run_update(message)
            return True
        except:
            logger.exception('Unexpected error in update')
            return False


    def try_sleep(self, start):
        # sleep for 10 ms but constantly check if
        # service would like to stop
        if self._interval is None:
            return

        elapsed = time.time_ns() - start
        sleep_time = self._interval - elapsed
        sleep_time = 0 if sleep_time < 0 else sleep_time
        start_sleep = time.time_ns()
        while time.time_ns() - start_sleep <= sleep_time:
            self.try_update()
            if self._stop_event.is_set():
                break
            time.sleep(0.01)


    def try_end(self):
        try:
            self.run_end()
        except:
            logger.exception('Unexpected error in end')
        self._stopped = True


    def run(self):
        if not self.try_start():
            return

        while not self._stop_event.is_set():
            start = time.time_ns()
            self.try_loop()
            self.try_update()
            self.try_sleep(start)
        self.try_end()


    def update(self, message):
        with self._update_lock:
            self._update_messages.append(message)


    def stop(self):
        self._stop_event.set()

    
    def stop_wait(self):
        self._stop_event.set()
        while not self.is_stopped():
            time.sleep(0.01)


    def is_stopped(self):
        return self._stopped


    def has_started(self):
        return self._has_started


import os
import time
import json
import threading
import logging

logger = logging.getLogger(__name__)

settings = {}
services = {}
triggers = {}

_setting_lock = threading.Lock()


def register_service(service_name, instance):
  global services
  if service_name in services:
    raise Exception(f'{service_name} already exists in services')

  if service_name == 'global':
    raise Exception(f'global is a reserved service name')
  
  services[service_name] = {
    'lock': threading.Lock(),
    'create': instance,
    'instance': None,
    'settings': {},
  }


def register_trigger(trigger_name, instance):
  global triggers
  if trigger_name in triggers:
    raise Exception(f'{trigger_name} already exists in triggers')
  
  triggers[trigger_name] = {
    'lock': threading.Lock(),
    'create': instance,
    'instance': None,
    'settings': {},
  }


def get_services_status():
  global services
  status = {}
  for service_name in services:
    status[service_name] = False
    with services[service_name]['lock']:
      if services[service_name]['instance'] is not None:
        if not services[service_name]['instance'].is_stopped():
          status[service_name] = True
  return status


def start_services():
  global services
  for service_name in services:
    start_service(service_name)


def start_service(service_name):
  global services
  logger.info(f'Start service {service_name}')
  if service_name in services:
    with services[service_name]['lock']:
      if services[service_name]['instance'] is None:

        if services[service_name]['settings'].get('disabled'):
          logger.info(f'Service {service_name} is disabled')
          return

        services[service_name]['instance'] = services[service_name]['create']()
        services[service_name]['instance'].start()


def start_triggers():
  global triggers
  for trigger_name in triggers:
    start_trigger(trigger_name)
  

def start_trigger(trigger_name):
  global triggers
  logger.info(f'Start trigger {trigger_name}')
  if trigger_name in triggers:
    with triggers[trigger_name]['lock']:
      if triggers[trigger_name]['instance'] is None:

        if triggers[trigger_name]['settings'].get('disabled'):
          logger.info(f'Trigger {trigger_name} is disabled')
          return

        triggers[trigger_name]['instance'] = triggers[trigger_name]['create']()
        triggers[trigger_name]['instance'].start()


def stop_triggers():
  global triggers
  for trigger_name in triggers:
    with triggers[trigger_name]['lock']:
      logger.info(f'Stop trigger {trigger_name}')
      if triggers[trigger_name]['instance'] is not None:
        triggers[trigger_name]['instance'].stop()
  for trigger_name in triggers:
    with triggers[trigger_name]['lock']:
      if triggers[trigger_name]['instance'] is not None:
        while not triggers[trigger_name]['instance'].is_stopped():
          time.sleep(0.01)
      triggers[trigger_name]['instance'] = None
      logger.info(f'Trigger {trigger_name} stopped!')


def stop_services():
  global services
  for service_name in services:
    with services[service_name]['lock']:
      logger.info(f'Stop service {service_name}')
      if services[service_name]['instance'] is not None:
        services[service_name]['instance'].stop()
  for service_name in services:
    with services[service_name]['lock']:
      if services[service_name]['instance'] is not None:
        while not services[service_name]['instance'].is_stopped():
          time.sleep(0.01)
      services[service_name]['instance'] = None
      logger.info(f'Service {service_name} stopped!')


def stop_service(service_name):
  global services
  logger.info(f'Stop service {service_name}')
  if service_name in services:
    with services[service_name]['lock']:
      if services[service_name]['instance'] is not None:
        services[service_name]['instance'].stop()
        while not services[service_name]['instance'].is_stopped():
          time.sleep(0.01)
        services[service_name]['instance'] = None
        logger.info(f'Service {service_name} stopped!')


def update_service(service_name, message):
  global services
  logger.info(f'Update service {service_name}: {message}')
  if service_name in services:
    with services[service_name]['lock']:
      if services[service_name]['instance'] is not None:
        services[service_name]['instance'].update(message)


def update_trigger(trigger_name, message=None):
  global triggers
  logger.info(f'Update trigger {trigger_name}')
  if trigger_name in triggers:
    with triggers[trigger_name]['lock']:
      if triggers[trigger_name]['instance'] is not None:
        triggers[trigger_name]['instance'].trigger(message)


def set_global_setting(key, value):
  global settings
  with _setting_lock:
    settings[key] = value


def get_global_setting(key, default=None):
  global settings
  with _setting_lock:
    return settings.get(key, default)


def set_service_setting(service_name, key, value):
  global services
  if service_name in services:
    with services[service_name]['lock']:
      services[service_name]['settings'][key] = value


def get_service_setting(service_name, key, default=None):
  global services
  if service_name in services:
    with services[service_name]['lock']:
      return services[service_name]['settings'].get(key, default)


def get_service_settings(service_name):
  global services
  if service_name in services:
    with services[service_name]['lock']:
      return services[service_name]['settings']

  
def get_all_settings():
  global services, settings
  with _setting_lock:
    returned = { 'global': settings }
  for service_name in services:
    with services[service_name]['lock']:
      returned[service_name] = services[service_name].get('settings', {})
  return returned


from setuptools import setup
from fservice import __version__

setup(
    name='fservice',
    version=__version__,    
    description='Stateful service management system',
    url='https://github.com/freefarmdata/fservice',
    author='Jack Mead',
    author_email='jackmead515@gmail.com',
    license='BSD 2-clause',
    packages=['fservice'],
    install_requires=[],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',  
        'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)


import threading
import logging
import time
from collections import deque

logger = logging.getLogger(__name__)

class Trigger(threading.Thread):

    def __init__(self, name='trigger'):
        super().__init__(name=name, daemon=True)
        self._stop_event = threading.Event()
        self._stopped = False
        self._has_started = False
        self._block_time = 5
        self._trigger = threading.Semaphore(0)
        self._queue = deque(maxlen=100000)


    def run_start(self):
        pass


    def run_trigger(self, message):
        pass

    
    def run_end(self):
        pass


    def try_start(self):
        self._has_started = True
        try:
            self.run_start()
            return True
        except:
            logger.exception('Unexpected error in start')
            self._stopped = True
            return False
    

    def try_end(self):
        try:
            self.run_end()
        except:
            logger.exception('Unexpected error in end')
        self._stopped = True


    def try_trigger(self):
        try:
            self.run_trigger(self._queue.pop())
            return True
        except:
            logger.exception('Unexpected error in trigger')
            return False


    def run(self):
        if not self.try_start():
            return

        while not self._stop_event.is_set():
            triggered = self._trigger.acquire(timeout=self._block_time)
            if triggered:
                self.try_trigger()
        self.try_end()


    def set_timeout(self, timeout=5):
        self._block_time = timeout


    def stop(self):
        self._stop_event.set()


    def trigger(self, message=None):
        self._queue.appendleft(message)
        self._trigger.release()


    def is_stopped(self):
        return self._stopped


    def has_started(self):
        return self._has_started